"""Special module that runs when the package is invoked directly
 from the command line using the -m flag.

For more information, see:
https://docs.python.org/3/library/__main__#main-py-in-python-packages.

All CLI entry points should be initialized in this module.
"""

# Copyright (C) 2025  Stefano Cuizza

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


from argparse import ArgumentParser, Namespace
import logging
from typing import Type

from . import cli
from .modules import log
from .parsers import MainArgumentParser


logger = logging.getLogger(__name__)


def init_cli_with_argument_parser(argument_parser: Type[ArgumentParser]) -> Namespace:
    """Initializes everything that is required by a Command Line Interface.

    The ArgumentParser class will be used to parse the command line parsing.

    Args:
        argument_parser:
            an **uninitialized** ArgumentParser class.

    Returns:
        the parsed command line parsing as a Namespace object.
    """

    log.init()

    with log.deferred_logging():
        log.set_root_logger_level(logging.INFO)

        # calling the ArgumentParser under deferred_logging is necessary to make
        # sure that the cached log records are handled with the correct logging
        # level.
        return argument_parser().parse_args()


def main() -> None:
    """Main CLI entry point."""

    arguments: Namespace = init_cli_with_argument_parser(MainArgumentParser)
    cli.main(arguments)


if __name__ == '__main__':
    main()



"""Contains logging helper functions and custom handlers/formatters."""

# Copyright (C) 2025  Stefano Cuizza

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


from collections import deque
from collections.abc import Callable
from contextlib import contextmanager
import logging
from sys import stdout
from typing import override

from .ansi import Fore, Style


class LoggingFormatter(logging.Formatter):
    """Base logging formatter that supports colorful output.

    Attributes:
        record_format_strings:
          dict mapping logging levels to their corresponding format strings.
    """

    record_format_strings: dict[int, str] = {
        logging.NOTSET:
            '%(message)s',

        logging.DEBUG:
            '[DEBUG] %(name)s:%(lineno)s %(message)s',

        logging.INFO:
            f'{Fore.LIGHTWHITE_EX}'
            f'[INFO] %(message)s'
            f'{Fore.RESET}',

        logging.WARNING:
            f'{Fore.YELLOW}'
            f'[WARNING] %(message)s'
            f'{Fore.RESET}',

        logging.ERROR:
            f'{Fore.LIGHTRED_EX}'
            f'[{Style.ITALIC}ERROR{Style.RESET_ITALIC}] '
            f'%(message)s'
            f'{Fore.RESET}',

        logging.CRITICAL:
            f'{Fore.LIGHTMAGENTA_EX}'
            f'[{Style.ITALIC}CRITICAL{Style.RESET_ITALIC}] '
            f'{Style.UNDERLINE}%(message)s{Style.RESET_UNDERLINE}'
            f'{Fore.RESET}',
    }

    @override
    def format(self, record: logging.LogRecord) -> str:
        """Formats the log record as text.

        Args:
            record:
                logging record to format.
        """

        fmt = self.record_format_strings.get(
            record.levelno,
            '%(message)s',
        )

        return logging.Formatter(fmt).format(record)


def init() -> None:
    """Initializes the package logger."""

    root = logging.getLogger()

    console_stream_handler = logging.StreamHandler(
        stream=stdout,
    )
    console_stream_handler.setLevel(logging.NOTSET)
    console_stream_handler.setFormatter(LoggingFormatter())

    root.addHandler(console_stream_handler)
    root.setLevel(logging.NOTSET)


def set_root_logger_level(level: int | str) -> None:
    """Changes the root logger logging level.

    Args:
        level:
            a valid logging level.
    """

    logging.getLogger().setLevel(level)
    for handler in logging.getLogger().handlers:
        handler.setLevel(level)


@contextmanager
def apply_filter(logging_filter: Callable[[logging.LogRecord], bool]):
    """Context manager that applies a filter to the root logger.

    Args:
        logging_filter:
            a logging filter (https://docs.python.org/3/library/logging.html#filter-objects).

    """

    logging.getLogger().addFilter(logging_filter)
    for handler in logging.getLogger().handlers:
        handler.addFilter(logging_filter)

    try:
        yield None
    finally:
        logging.getLogger().removeFilter(logging_filter)
        for handler in logging.getLogger().handlers:
            handler.removeFilter(logging_filter)


@contextmanager
def deferred_logging():
    """Context manager that caches logging calls and handles them all on exit."""

    cached_records = deque()

    def cache_filter(record: logging.LogRecord) -> bool:
        """"""

        cached_records.append(record)
        return False

    try:
        with apply_filter(cache_filter):
            yield None
    finally:
        for record in cached_records:
            logging.getLogger().handle(record)



"""Adds new Sphinx directives and features to manage and display package metadata.

The Python package from which the metadata is pulled is the one whose name is
the same as the value of the 'project' variable defined in the namespace of the
project’s configuration.


New directives:
    - .. projecturls::
      Adds a toc tree full of project urls extracted from the Project-URL
      entry in the package metadata.

    - .. projectstatusalert::
      Displays an admonition if the development status of the Python
      package (extracted from the package classifiers) is equal or below a
      certain development stage.

      For more information on development status classifiers, see:
      https://pypi.org/classifiers.


New replace directives:
    - |author|
      the package author.
    - |license|
      the package license (e.g. Creative Commons).
    - |summary|
      the package summary.


For more information on package metadata, see:
https://docs.python.org/3/library/importlib.metadata.html#overview.
"""

import logging
from typing import Any

from docutils import nodes
from docutils.nodes import Node
from docutils.parsers.rst import directives
from docutils.statemachine import StringList
from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.util.docutils import SphinxDirective
from sphinx.util.typing import OptionSpec


type PackageMetadata = dict[str, str | list[str]]

logger = logging.getLogger(__name__)


# taken from: https://pypi.org/classifiers/.
DEV_STATUS_MAPPING: dict[int, tuple[str]] = {
    1: ('Planning',),
    2: ('Pre-Alpha',),
    3: ('Alpha',),
    4: ('Beta',),
    5: ('Production/Stable', 'Production', 'Stable'),
    6: ('Mature',),
    7: ('Inactive',),
}


class ProjectURLNode(nodes.Element):
    pass


class DevStatusAlertNode(nodes.Element):
    pass


class ProjectURLsDirective(SphinxDirective):
    """Sphinx directive that add a toc tree of project urls
    extracted from the specified Python package metadata.

    Args:
        caption:
          optional caption for the toc tree.
        hidden:
          optional flag to hide the toc tree.
    """

    app: Sphinx

    has_content: bool = True

    optional_arguments = 1
    required_arguments = 0

    option_spec: OptionSpec = {
        'caption': directives.unchanged_required,
        'hidden': directives.flag,
    }

    def run(self) -> list[Node]:
        """Sphinx Directive entry point."""

        if self.env.docname != self.env.config.master_doc:
            return []

        body, indent = StringList(['.. toctree::']), ' ' * 4

        if 'caption' in self.options:
            body.append(StringList([f'{indent}:caption: {self.options['caption']}']))

        if 'hidden' in self.options:
            body.append(StringList([f'{indent}:hidden:']))

        body.append(StringList([f'{indent}\n']))
        for website_name, website_url in (
                self.app.config['metadata_project_urls']
                if 'metadata_project_urls' in self.app.config
                else {}
        ).items():
            body.append(StringList([f'{indent}{website_name} <{website_url}>']))

        for line in self.content:
            body.append(StringList([f'{indent}{line}']))

        body.append(StringList(['\n', '\n']))

        node = addnodes.only(expr='html')
        content_node = nodes.paragraph(rawsource=str(body))
        node += content_node
        self.state.nested_parse(
            StringList(body),
            self.content_offset,
            content_node,
        )

        return [node]


class DevStatusAlertDirective(SphinxDirective):
    """Displays an admonition if the development status of the Python
      package (extracted from the package classifiers) is equal or below a
      certain development stage.

      Options:
        :title:
          optional admonition title.
          default title: Development Status
        :body:
          optional admonition body. The strings %(index)d and %(literal)s will
          be replaced by the package development status index and literal
          string (e.g. alpha, beta, ...).
        :status:
          the threshold status. can either be a number or a string.
    """

    optional_arguments = 2
    required_arguments = 1

    option_spec: OptionSpec = {
        'title': directives.unchanged,
        'body': directives.unchanged,
    }

    def run(self) -> list[Node]:
        """Sphinx Directive entry point."""

        node = DevStatusAlertNode()
        node.document = self.state.document
        self.set_source_info(node)

        if title := self.options.get('title'):
            node['title'] = title

        if body := self.options.get('body'):
            node['body'] = body

        node['development_status'] = self.arguments[0]

        return [node]


def process_devstatusalert_nodes(app: Sphinx, doctree: nodes.document, docname: str) -> None:
    """Converts DevStatusAlertNodes into admonitions warning the user if the
    package status is below or equal to a certain development stage.

    The package from which the project urls are extracted is the one whose name
    is the same as the value of the 'project' variable defined in the namespace
    of the project’s configuration.

    Args:
        app:
          Sphinx instance.
        doctree:
          the doctree of the file.
        docname:
          the name of the source file.
    """

    project_development_status = (
        app.config['metadata_project_development_status']
    )

    for node in list(doctree.findall(DevStatusAlertNode)):
        if 'development_status' not in node:
            node.replace_self([])
            continue

        try:
            status_index = int(node['development_status'])

        except ValueError:
            status_literal = node['development_status']

            status_index = None
            for index, literals in DEV_STATUS_MAPPING.items():
                if status_literal.lower() in map(str.lower, literals):
                    status_index = index
                    break

            if status_index is None:
                logger.warning(
                    '[%s] development status literal %r does '
                    'not map to any development status index.',
                    __name__, status_literal,
                )

                node.replace_self([])
                return

        if project_development_status[0] > status_index:
            node.replace_self([])
            continue

        admonition_node = nodes.admonition('')
        admonition_node += nodes.title(
            text=node.get('title', 'Development Status') % {
                 'index': project_development_status[0],
                 'literal': project_development_status[1].lower(),
             }
        )
        admonition_node += nodes.paragraph(
            text=node.get(
                'body',
                'Lorem ipsum dolor sit amet, '
                'consectetur adipiscing elit, '
                'sed do eiusmod tempor incididunt '
                'ut labore et dolore magna aliqua.'
            ) % {
                 'index': project_development_status[0],
                 'literal': project_development_status[1].lower(),
             }

        )

        node.replace_self([admonition_node])


def add_global_rep(app: Sphinx, docname: str, source: list[str]) -> None:
    """Adds some special replace directives at the end of every source file,
    and makes sure that the default ones (e.g. project) work as intended.

    - |author|:
        the package author.
    - |license|:
        the package license (e.g. Creative Commons).
    - |summary|:
        the package summary string.

    Args:
        app:
          Sphinx instance.
        docname:
          the name of the source file.
        source:
          a list whose single element is the contents of the source file.
    """

    app_config: dict = {
        config_value.name: config_value.value for config_value in app.config
    }

    default_repl_fix = '\n'.join((
        f'.. |project| replace:: {app_config.get('project', '|project|')}',
        f'.. |release| replace:: {app_config.get('release', '|release|')}',
        f'.. |version| replace:: {app_config.get('version', '|version|')}',
    ))

    custom_rst_epilog = '\n'.join((
        f'.. |author| replace:: {app_config.get('author', '|author|')}',
        f'.. |license| replace:: {app_config.get('metadata_project_license_type', '|license|')}',
        f'.. |summary| replace:: {app_config.get('metadata_project_json', {}).get('summary', '|summary|')}',
        '    :lower:',
    ))

    repl_source = '\n'.join((
        *source,
        default_repl_fix,
        custom_rst_epilog,
    )) + '\n'

    source.insert(0, repl_source)
    source.pop()


def setup(app: Sphinx) -> dict[str, Any]:
    """Sphinx extension entry point.

    Args:
        app:
          Sphinx instance.

    Returns:
        the metadata of the extension.
    """

    pud = ProjectURLsDirective
    pud.app = app

    app.setup_extension('metadata')
    app.add_directive('projecturls', ProjectURLsDirective)

    app.add_node(DevStatusAlertNode)
    app.add_directive('devstatusalert', DevStatusAlertDirective)
    app.connect('doctree-resolved', process_devstatusalert_nodes)

    app.connect('source-read', add_global_rep)

    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



from importlib import import_module
from importlib.metadata import metadata as importlib_metadata
from pathlib import Path
import tomllib

import pytest


def get_package_name(pyproject_toml: Path) -> str:
    """Returns the package name extracted from the pyproject.toml file.

    Args:
        pyproject_toml:
          path to the pyproject.toml config file.
    """

    with open(pyproject_toml, 'rb') as f:
        data = tomllib.load(f)

    return data['project']['name']


def get_package_metadata(package_name: str) -> dict:
    """Returns the JSON representation of the package metadata.

    For more information, see:
    https://docs.python.org/3/library/importlib.metadata.html#distribution-metadata.

    Args:
        package_name:
          the name of the package from which to extract the metadata.
    """

    return importlib_metadata(package_name).json


@pytest.fixture(autouse=True, scope='session')
def init_global_variables():
    """Initializes global variables used for testing purposed."""

    package_pyproject_path: Path = Path.resolve(
        Path(__file__).parent / '..' / 'pyproject.toml'
    )
    package_name: str = get_package_name(package_pyproject_path)

    pytest.PACKAGE_MODULE = import_module(package_name)
    pytest.PACKAGE_METADATA = get_package_metadata(package_name)


@pytest.fixture(scope='session')
def package():
    """Returns the package that is being tested."""

    return pytest.PACKAGE_MODULE


@pytest.fixture(scope='session')
def metadata():
    """Returns the package metadata."""

    return pytest.PACKAGE_METADATA



"""Contains argument types."""

# Copyright (C) 2025  Stefano Cuizza

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import Literal

from jinja2 import Template

from .. import tokenizer
from ..arduino.keyboard import LAYOUTS
from ..modules import template as template_
from ..tokenizer import RawKeystroke


def key(argument: str) -> list[RawKeystroke]:
    """A series of keystrokes."""

    special_characters_mapping: dict[str, str] = {
        '\\t': '\t',
        '\\n': '\n',
        '\\r': '\r',
        '\\f': '\f',
        '\\v': '\v'
    }

    # fix misinterpreted special characters.
    for character, replacement in special_characters_mapping.items():
        argument = argument.replace(character, replacement)

    return tokenizer.to_keystrokes(argument)


def delay(argument: str) -> int:
    """Delay in milliseconds, strictly greater than 0."""


    if (ms := int(argument)) <= 0:
        raise ValueError('delay must be greater than 0')

    return ms


def key_or_delay(argument: str) -> list[int | RawKeystroke]:
    """A series of keystrokes and delays."""

    if argument.isnumeric():
        return [delay(argument)]

    return key(argument)


def keyboard_layout(argument: str) -> str:
    """A supported keyboard layout."""

    try:
        _constant = LAYOUTS[argument]
    except KeyError as e:
        raise ValueError from e

    return argument


def template(argument: str) -> Template:
    """A valid sketch template."""

    return template_.get(argument)


def repeat_every(argument: Literal[False] | str) -> Literal[False] | int:
    """Optionally repeat the injection every n milliseconds."""

    if isinstance(argument, bool) and argument is False:
        return argument

    return int(argument)


def words_per_minute(argument: str) -> int:
    """The speed at which keystrokes are sent, in words per minute."""

    if (wpm := int(argument)) < 0:
        raise ValueError('wpm must be a positive number')

    return wpm



"""Provides easy access to the package metadata."""

# Copyright (C) 2025  Stefano Cuizza

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


from importlib import resources
from pathlib import Path
import shutil
from typing import Literal, TypedDict

from jinja2 import Environment, FileSystemLoader, Template

from ..modules import metadata
from ..tokenizer import Keystroke, SleepInterval


TEMPLATES_DIRECTORY: Path = Path(
    str(resources.files(metadata.package()) / 'templates')
)
ENVIRONMENT = Environment(
    loader=FileSystemLoader(TEMPLATES_DIRECTORY)
)


class TemplateParameters(TypedDict):
    """Parameters accepted by a template."""

    do_repeat: Literal[False] | int
    interval_mapping: list[tuple[int, SleepInterval]]
    keyboard_layout: str
    keystrokes: list[Keystroke]
    wpm: int


def add(path: Path) -> None:
    """Adds a template to the package collection.

    Args:
        path:
            the template's path.
    """

    shutil.copy(path, TEMPLATES_DIRECTORY)


def enum() -> list[str]:
    """Lists all available templates."""

    return ENVIRONMENT.list_templates()


def get(name: str) -> Template:
    """Loads a template by name.

    Args:
        name:
            the name of the template.
    """

    return ENVIRONMENT.get_template(name)


def remove(name: str) -> None:
    """Removes a template from the package's collection.

    Args:
        name:
            the name of the template to remove.
    """

    (TEMPLATES_DIRECTORY / name).unlink()



"""Provides easy access to the package metadata by adding new configuration
values to app.config.

The package from which the metadata is extracted is the one whose name
is the same as the value of the 'project' variable defined in the namespace
of the project’s configuration.

New configuration values:
    - metadata_project_json:
      contains the package metadata as a JSON dictionary.
    - metadata_project_development_status:
      contains the package development status and its associated stage index,
      extracted from the "Development Status" classifier (if present).
    - metadata_project_license_type:
      contains the package license type, extracted from the "License" classifier
      (if present).
    - metadata_project_urls:
      contains a dict mapping the project websites with their corresponding urls.
"""

from importlib.metadata import metadata, PackageNotFoundError
import logging
from operator import methodcaller
from typing import Any

from sphinx.application import Sphinx


type JSONPackageMetadata = dict[str, str | list[str]]

logger = logging.getLogger(__name__)


def get_package_metadata(package_name: str) -> JSONPackageMetadata:
    """Returns the package metadata in JSON format.

    For more information, see:
    https://docs.python.org/3/library/importlib.metadata.html#distribution-metadata.

    Args:
        package_name:
          the name of the package.

    Raises:
        PackageNotFoundError:
          the specified package was not found.
    """

    try:
        package_metadata: JSONPackageMetadata = (
            metadata(package_name).json
        )
    except PackageNotFoundError:
        logger.critical('%r package not found.', package_name)
        raise

    return package_metadata


def get_package_development_status(package_metadata: JSONPackageMetadata) -> tuple[int, str]:
    """Returns the package development status and its associated stage index.

    The status is extracted from the "Development Status" classifier (if present).

    Common possible return values:
        - 1, Planning
        - 2, Pre-Alpha
        - 3, Alpha
        - 4, Beta
        - 5, Production/Stable
        - 6, Mature
        - 7, Inactive

    Args:
        package_metadata:
          package metadata dump in JSON format.
    """

    for classifier_type, classifier_body in map(
            methodcaller('split', ' :: ', 1),
            package_metadata['classifier'],
    ):
        if classifier_type == 'Development Status':
            status_number, status_literal = classifier_body.split(' - ')

            return int(status_number), status_literal

    return 0, ''


def get_package_license_type(package_metadata: JSONPackageMetadata) -> str:
    """Returns the package license type.

    The package license type is extracted from the "License" classifier (if present).

    Args:
        package_metadata:
          package metadata dump in JSON format.
    """

    for classifier_type, classifier_body in map(
            methodcaller('split', ' :: ', 1),
            package_metadata['classifier'],
    ):
        if classifier_type == 'License':
            return classifier_body.split(' :: ')[-1]

    return ''


def get_package_urls(package_metadata: JSONPackageMetadata) -> dict[str, str]:
    """Returns a dict mapping the project websites with their corresponding urls.

    Args:
        package_metadata:
          package metadata dump in JSON format.
    """

    package_project_url_field: list[str] = (
        package_metadata['project_url']
    )

    package_project_urls: dict[str, str] = {}

    # Raymond Hettinger (2011, December 11). python map (...). StackOverflow.
    # https://stackoverflow.com/a/8461254.
    for website_name, website_url in map(
            methodcaller('split', ', '),
            package_project_url_field,
    ):
        package_project_urls[website_name] = website_url

    return package_project_urls


def setup(app: Sphinx) -> dict[str, Any]:
    """Sphinx extension entry point.

    Args:
        app:
          Sphinx instance.

    Returns:
        the metadata of the extension.
    """

    app_config: dict = {
        config_value.name: config_value.value for config_value in app.config
    }
    package_name: str = app_config['project']
    package_metadata: JSONPackageMetadata = get_package_metadata(package_name)

    app.add_config_value(
        name='metadata_project_json',
        default=package_metadata,
        rebuild='',
        types=JSONPackageMetadata.__value__,
    )

    app.add_config_value(
        name='metadata_project_development_status',
        default=get_package_development_status(package_metadata),
        rebuild='',
        types=list,
    )

    app.add_config_value(
        name='metadata_project_license_type',
        default=get_package_license_type(package_metadata),
        rebuild='',
        types=str,
    )

    app.add_config_value(
        name='metadata_project_urls',
        default=get_package_urls(package_metadata),
        rebuild='',
        types=dict[str, str],
    )

    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



"""Fixes the :caption: option in the autosummary directive by monkeypatching
the run method of the Autosummary class.

Problem:
    For some reason, when a caption is specified, it is not actually rendered.

    There is a GitHub issue that mentions this:
    https://github.com/sphinx-doc/sphinx/issues/7490.
"""

from typing import Any, Callable

from docutils import nodes
from docutils.nodes import Node
from sphinx.application import Sphinx
from sphinx.ext import autosummary


def patch_autosummary_directive_run_method(
        func: Callable[[autosummary.Autosummary], list[Node]]
) -> Callable[[autosummary.Autosummary], list[Node]]:
    """Decorator function to patch the run method of the Autosummary directive."""

    def patch(self: autosummary.Autosummary) -> list[Node]:
        nodes_list = func(self)

        if caption := self.options.get('caption'):
            section_node = nodes.section(ids=[nodes.make_id(caption)])
            section_node += nodes.title(text=caption)
            section_node += nodes_list

            nodes_list = [section_node]

        return nodes_list

    if hasattr(func, '__doc__'):
        patch.__doc__ = func.__doc__

    return patch


def apply_autosummary_directive_patch(
        cls: [autosummary.Autosummary],
) -> autosummary.Autosummary:
    """Applies the Autosummary directive patch."""

    cls.run = patch_autosummary_directive_run_method(
        cls.run
    )

    return cls


def setup(app: Sphinx) -> dict[str, Any]:
    """Sphinx extension entry point.

    Args:
        app:
          Sphinx instance.

    Returns:
        the metadata of the extension.
    """

    app.setup_extension('sphinx.ext.autosummary')
    apply_autosummary_directive_patch(autosummary.Autosummary)

    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



"""Adds the ability to extract the name of the module to be used in the
autosummary directive from a configuration value.

New options:
    - :fromconfig: config
      the package module name will be extracted fromm the config value.


Typical usage:

    .. autosummary::
        :toctree: _autosummary
        :fromconfig: project

    Will produce a summary of the package named after the string contained
    in the `project` variable inside the conf.py file.
"""

import re
from typing import Any, Callable, Optional

from sphinx.ext import autosummary
from docutils.nodes import Node
from docutils.parsers.rst import directives
from sphinx.application import Sphinx
from sphinx.ext.autosummary.generate import AutosummaryEntry


def patch_autosummary_directive_run_method(
        app: Sphinx,
        func: Callable[[autosummary.Autosummary], list[Node]]
) -> Callable[[autosummary.Autosummary], list[Node]]:
    """Decorator function to patch the run method of the Autosummary directive."""

    def patch(self: autosummary.Autosummary) -> list[Node]:
        """"""

        app_config: dict = {
            config_value.name: config_value.value for config_value in app.config
        }

        if 'fromconfig' in self.options:
            if package_name := app_config.get(self.options['fromconfig'], None):
                self.content = [package_name]
                self.block_text = self.block_text + f'{' ' * 4}tool\n\n'
        output = func(self)
        return output

    if hasattr(func, '__doc__'):
        patch.__doc__ = func.__doc__

    return patch


def patch_find_autosummary_in_lines_function(
        app: Sphinx,
        func: Callable[[list[str], Optional[str], Optional[str]], list[AutosummaryEntry]]
) -> Callable[[list[str], Optional[str], Optional[str]], list[AutosummaryEntry]]:
    """Decorator function to patch the find_autosummary_in_lines function
    of the autosummary.generate module.
    """

    indent: str = ' ' * 4
    app_config: dict = {
        config_value.name: config_value.value for config_value in app.config
    }

    def patch(
            lines: list[str],
            module: Optional[str] = None,
            filename: Optional[str] = None,
    ) -> list[AutosummaryEntry]:
        fromconfig_arg_re = re.compile(r'^\s+:fromconfig:\s*(.*?)\s*$')

        for i, line in enumerate(lines):
            if m := fromconfig_arg_re.match(line):
                conf_value = m.group(1).strip()
                if package_name := app_config.get(conf_value, None):
                    lines[i] = f'{indent}{package_name}\n'

        return func(lines, module, filename)

    if hasattr(func, '__doc__'):
        patch.__doc__ = func.__doc__

    return patch


def apply_autosummary_directive_patch(
        app: Sphinx,
        cls: [autosummary.Autosummary],
) -> autosummary.Autosummary:
    """Applies the Autosummary directive patch."""

    if not cls.option_spec:
        cls.option_spec = {}

    cls.option_spec |= {
        'fromconfig': directives.unchanged,
    }

    cls.run = patch_autosummary_directive_run_method(app, cls.run)

    return cls


def apply_autosummary_function_patch(
        app: Sphinx,
        module: [autosummary.generate],
) -> autosummary.generate:
    """Applies the autosummary.generate.find_autosummary_in_lines function patch."""

    module.find_autosummary_in_lines = patch_find_autosummary_in_lines_function(
        app,
        module.find_autosummary_in_lines,
    )

    return module


def setup(app: Sphinx) -> dict[str, Any]:
    """Sphinx extension entry point.

    Args:
        app:
          Sphinx instance.

    Returns:
        the metadata of the extension.
    """

    app.setup_extension('sphinx.ext.autosummary')

    autosummary.Autosummary = apply_autosummary_directive_patch(
        app,
        autosummary.Autosummary,
    )
    autosummary.generate = apply_autosummary_function_patch(
        app,
        autosummary.generate,
    )

    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



"""Contains keyboard library constants."""

# Copyright (C) 2025  Stefano Cuizza

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


# https://docs.arduino.cc/language-reference/en/functions/usb/Keyboard/keyboardBegin/#Keyboard%20layouts.
LAYOUTS: dict[str, str] = {
    'dk': 'KeyboardLayout_da_DK',
    'de': 'KeyboardLayout_de_DE',
    'us': 'KeyboardLayout_en_US',
    'es': 'KeyboardLayout_es_ES',
    'fr': 'KeyboardLayout_fr_FR',
    'hu': 'KeyboardLayout_hu_HU',
    'it': 'KeyboardLayout_it_IT',
    'pt': 'KeyboardLayout_pt_PT',
    'se': 'KeyboardLayout_sv_SE',
}

# https://docs.arduino.cc/language-reference/en/functions/usb/Keyboard/keyboardModifiers/.
SPECIAL_KEYS: dict[str, str] = {
    # keyboard modifiers.
    'LEFT_CTRL': 'KEY_LEFT_CTRL',
    'LEFT_SHIFT': 'KEY_LEFT_SHIFT',
    'LEFT_ALT': 'KEY_LEFT_ALT',
    'LEFT_GUI': 'KEY_LEFT_GUI',
    'RIGHT_CTRL': 'KEY_RIGHT_CTRL',
    'RIGHT_SHIFT': 'KEY_RIGHT_SHIFT',
    'RIGHT_ALT': 'KEY_RIGHT_ALT',
    'RIGHT_GUI': 'KEY_RIGHT_GUI',

    # within the alphanumeric cluster.
    'TAB': 'KEY_TAB',
    'CAPS_LOCK': 'KEY_CAPS_LOCK',
    'BACKSPACE': 'KEY_BACKSPACE',
    'RETURN': 'KEY_RETURN',
    'MENU': 'KEY_MENU',

    # navigation cluster.
    'INSERT': 'KEY_INSERT',
    'DELETE': 'KEY_DELETE',
    'HOME': 'KEY_HOME',
    'END': 'KEY_END',
    'PAGE_UP': 'KEY_PAGE_UP',
    'PAGE_DOWN': 'KEY_PAGE_DOWN',
    'UP_ARROW': 'KEY_UP_ARROW',
    'DOWN_ARROW': 'KEY_DOWN_ARROW',
    'LEFT_ARROW': 'KEY_LEFT_ARROW',
    'RIGHT_ARROW': 'KEY_RIGHT_ARROW',

    # numeric keypad.
    'NUM_LOCK': 'KEY_NUM_LOCK',
    'KP_SLASH': 'KEY_KP_SLASH',
    'KP_ASTERISK': 'KEY_KP_ASTERISK',
    'KP_MINUS': 'KEY_KP_MINUS',
    'KP_PLUS': 'KEY_KP_PLUS',
    'KP_ENTER': 'KEY_KP_ENTER',
    'KP_1': 'KEY_KP_1',
    'KP_2': 'KEY_KP_2',
    'KP_3': 'KEY_KP_3',
    'KP_4': 'KEY_KP_4',
    'KP_5': 'KEY_KP_5',
    'KP_6': 'KEY_KP_6',
    'KP_7': 'KEY_KP_7',
    'KP_8': 'KEY_KP_8',
    'KP_9': 'KEY_KP_9',
    'KP_0': 'KEY_KP_0',
    'KP_DOT': 'KEY_KP_DOT',

    # escape and function keys.
    'ESC': 'KEY_ESC',
    'F1': 'KEY_F1',
    'F2': 'KEY_F2',
    'F3': 'KEY_F3',
    'F4': 'KEY_F4',
    'F5': 'KEY_F5',
    'F6': 'KEY_F6',
    'F7': 'KEY_F7',
    'F8': 'KEY_F8',
    'F9': 'KEY_F9',
    'F10': 'KEY_F10',
    'F11': 'KEY_F11',
    'F12': 'KEY_F12',
    'F13': 'KEY_F13',
    'F14': 'KEY_F14',
    'F15': 'KEY_F15',
    'F16': 'KEY_F16',
    'F17': 'KEY_F17',
    'F18': 'KEY_F18',
    'F19': 'KEY_F19',
    'F20': 'KEY_F20',
    'F21': 'KEY_F21',
    'F22': 'KEY_F22',
    'F23': 'KEY_F23',
    'F24': 'KEY_F24',

    # function control keys.
    'PRINT_SCREEN': 'KEY_PRINT_SCREEN',
    'SCROLL_LOCK': 'KEY_SCROLL_LOCK',
    'PAUSE': 'KEY_PAUSE',
}



"""Adds new functionalities to the docutils replace directive.

New options:
    - :lower:
      converts the substitution reference to lowercase.

    - :upper:
      converts the substitution reference to uppercase.

    - :ltrim:
      whitespace to the left of the substitution reference is removed.

    - :rtrim:
      whitespace to the right of the substitution reference is removed.

    - :trim:
      equivalent to ltrim plus rtrim; whitespace on both sides of the substitution reference is removed.
"""

from typing import Any, Callable

from docutils.nodes import Node
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives import misc
from sphinx.application import Sphinx


def apply_replace_run_patch(
        func: Callable[[misc.Replace], list[Node]]
) -> Callable[[misc.Replace], list[Node]]:
    """Decorator function to patch the run method of the Replace directive."""

    def patch(self: misc.Replace) -> list[Node]:

        if 'lower' in self.options:
            self.content.data = list(map(str.lower, self.content.data))

        if 'upper' in self.options:
            self.content.data = list(map(str.upper, self.content.data))

        if 'ltrim' in self.options:
            self.content.data = list(map(str.ltrim, self.content.data))

        if 'rtrim' in self.options:
            self.content.data = list(map(str.rtrim, self.content.data))

        if 'trim' in self.options:
            self.content.data = list(map(str.trim, self.content.data))

        return func(self)

    if hasattr(func, '__doc__'):
        patch.__doc__ = func.__doc__

    return patch


def apply_replace_patch(
        cls: misc.Replace
) -> misc.Replace:
    """Decorator function to patch the Replace directive."""
    
    if not cls.option_spec:
        cls.option_spec = {}

    cls.option_spec |= {
        'lower': directives.flag,
        'upper': directives.flag,
        'trim': directives.flag,
        'ltrim': directives.flag,
        'rtrim': directives.flag,
    }

    cls.run = apply_replace_run_patch(cls.run)

    return cls


def setup(app: Sphinx) -> dict[str, Any]:
    """Sphinx extension entry point.

    Args:
        app:
          Sphinx instance.

    Returns:
        the metadata of the extension.
    """

    misc.Replace = apply_replace_patch(misc.Replace)

    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



"""Main argument parser module."""

# Copyright (C) 2025  Stefano Cuizza

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


from argparse import (
    ArgumentParser, FileType, Namespace,
)
from collections.abc import Sequence
import itertools
import logging
from typing import Optional, override

from ..arduino.keyboard import LAYOUTS
from ..modules import metadata, template
from ..modules.parsing.parsers import MainArgumentParserTemplate
from ..tokenizer import keystrokes_to_arrays, merge_delays
from . import types


logger = logging.getLogger(__name__)


def _construct() -> ArgumentParser:
    """Returns an instance of the module's argument parser.

    Invoked by the `argparse` directive in the docs.
    For more information, see https://sphinx-argparse.readthedocs.io/en/stable.
    """

    return MainArgumentParser()


class MainArgumentParser(MainArgumentParserTemplate):
    """Handles the arguments that get passed to the package when its invoked
    directly from the command line (python -m package_name ...).
    """

    @override
    def __init__(self):
        super().__init__(
            prog=metadata.package(),
            description=metadata.summary(),
            prefix_chars='-',
        )

    def _extend_arguments(self) -> None:
        default_keyboard_layout: str = 'us'
        default_outfile_name: str = 'sketch.ino'
        default_template_name: str = 'any.default'
        default_wpm: int = 180

        self.add_argument(
            'keystrokes',
            action='store',
            help='sequence of keystrokes with optional delays between them',
            metavar='string | delay',
            nargs='+',
            type=types.key_or_delay,
        )

        self.add_argument(
            '-l', '--layout',
            action='store',
            choices=LAYOUTS.keys(),
            default=default_keyboard_layout,
            help='keyboard layout (default: %(default)s)',
            type=types.keyboard_layout,
        )

        self.add_argument(
            '-t', '--template',
            action='store',
            choices=template.enum(),
            default=default_template_name,
            help=(
                'the sketch template to use to generate '
                'the payload (default: %(default)s)'
            ),
            type=types.template,
        )

        self.add_argument(
            '-w', '--wpm',
            action='store',
            default=default_wpm,
            help=(
                'the average speed at which characters are '
                'typed, in words per minute. A value of zero '
                'will add no delay between keystrokes '
                '(default: %(default)d)'
            ),
            type=types.words_per_minute,
        )

        self.add_argument(
            '-r', '--repeat',
            action='store',
            const=0,
            default=False,
            metavar='ms',
            nargs='?',
            help=(
                'repeat the injection in a loop. Optionally '
                'specify a delay between repetitions in milliseconds'
            ),
            type=types.repeat_every,
        )

        self.add_argument(
            '-o', '--outfile',
            action='store',
            default=default_outfile_name,
            help=f'the output file (default: {default_outfile_name})',
            metavar='path',
            type=FileType('w+', encoding='utf-8'),
        )

    def _extend_subparsers(self) -> None:
        pass

    @override
    def parse_args(
            self,
            args: Optional[Sequence[str]] = None,
            namespace: Optional[Namespace] = None
    ) -> Namespace:

        namespace = super().parse_args(args=args, namespace=namespace)

        keystrokes, interval_mapping = keystrokes_to_arrays(
            merge_delays(list(itertools.chain(*namespace.keystrokes)))
        )

        namespace.keystrokes = keystrokes
        namespace.interval_mapping = interval_mapping

        # inclusion in the choices sequence is checked after any type conversions have been performed.
        namespace.layout = LAYOUTS[namespace.layout]

        return namespace



"""Command Line Interface entry point."""

# Copyright (C) 2025  Stefano Cuizza

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


from argparse import Namespace
import logging

from jinja2 import Template

from .modules.template import TemplateParameters


logger = logging.getLogger(__name__)


def main(namespace: Namespace) -> None:
    """Main CLI function.

    Args:
        namespace:
          Namespace containing the command line parsing.
    """

    chosen_template: Template = namespace.template
    template_parameters: TemplateParameters = {
        'do_repeat': namespace.repeat,
        'interval_mapping': namespace.interval_mapping,
        'keyboard_layout': namespace.layout,
        'keystrokes': namespace.keystrokes,
        'wpm': namespace.wpm,
    }

    payload = chosen_template.render(**template_parameters)

    namespace.outfile.write(payload)
    namespace.outfile.close()



"""This module extends colorama's ANSI codes.

All AnsiType classes store only the ANSI integer codes, while Fore, Back, and Style store the formatted ANSI codes.

Typical usage example:

    print(f'{Fore.RED}helloworld!{Fore.RESET}')
"""


# Copyright (C) 2025  Stefano Cuizza

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import Final

from colorama.ansi import AnsiCodes


class AnsiFore(AnsiCodes):
    """Struct containing ANSI foreground int codes."""

    BLACK: Final[int] = 30
    RED: Final[int] = 31
    GREEN: Final[int] = 32
    YELLOW: Final[int] = 33
    BLUE: Final[int] = 34
    MAGENTA: Final[int] = 35
    CYAN: Final[int] = 36
    WHITE: Final[int] = 37

    LIGHTBLACK_EX: Final[int] = 90
    LIGHTRED_EX: Final[int] = 91
    LIGHTGREEN_EX: Final[int] = 92
    LIGHTYELLOW_EX: Final[int] = 93
    LIGHTBLUE_EX: Final[int] = 94
    LIGHTMAGENTA_EX: Final[int] = 95
    LIGHTCYAN_EX: Final[int] = 96
    LIGHTWHITE_EX: Final[int] = 97

    RESET: Final[int] = 39


class AnsiBack(AnsiCodes):
    """Struct containing ANSI background int codes."""

    BLACK: Final[int] = 40
    RED: Final[int] = 41
    GREEN: Final[int] = 42
    YELLOW: Final[int] = 43
    BLUE: Final[int] = 44
    MAGENTA: Final[int] = 45
    CYAN: Final[int] = 46
    WHITE: Final[int] = 47

    LIGHTBLACK_EX: Final[int] = 100
    LIGHTRED_EX: Final[int] = 101
    LIGHTGREEN_EX: Final[int] = 102
    LIGHTYELLOW_EX: Final[int] = 103
    LIGHTBLUE_EX: Final[int] = 104
    LIGHTMAGENTA_EX: Final[int] = 105
    LIGHTCYAN_EX: Final[int] = 106
    LIGHTWHITE_EX: Final[int] = 107

    RESET: Final[int] = 49


class AnsiStyle(AnsiCodes):
    """Struct containing ANSI style int codes."""

    BRIGHT: Final[int] = 1
    DIM: Final[int] = 2
    ITALIC: Final[int] = 3
    UNDERLINE: Final[int] = 4
    BLINKINGMODE: Final[int] = 5
    NEGATIVE: Final[int] = 7
    HIDDEN: Final[int] = 8
    STRIKETHROUGH: Final[int] = 9

    RESET_BRIGHT: Final[int] = 22
    RESET_DIM: Final[int] = 22
    RESET_ITALIC: Final[int] = 23
    RESET_UNDERLINE: Final[int] = 24
    RESET_BLINKINGMODE: Final[int] = 25
    RESET_NEGATIVE: Final[int] = 27
    RESET_HIDDEN: Final[int] = 28
    RESET_STRIKETHROUGH: Final[int] = 29
    RESET_ALL: Final[int] = 0


Fore = AnsiFore()
"""Struct containing ANSI foreground codes."""

Back = AnsiBack()
"""Struct containing ANSI background codes."""

Style = AnsiStyle()
"""Struct containing ANSI style codes."""



from datetime import datetime
from importlib import import_module
import os
from pathlib import Path
import sys
import tomllib

from sphinx.application import Sphinx


sys.path.append(os.path.abspath('./_ext'))


def get_package_name(pyproject_toml: Path) -> str:
    """Returns the project name contained in the pyproject.toml file.

    Args:
        pyproject_toml:
          path to the pyproject.toml configuration file.
    """

    with open(pyproject_toml, 'rb') as f:
        data = tomllib.load(f)

    return data['project']['name']


pyproject_path: Path = (
    Path.resolve(Path(__file__).parent / '..' / '..' / 'pyproject.toml')
)
package = import_module(get_package_name(pyproject_path))


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = package.__package__
author = package.__author__
release = package.__version__
version = '.'.join(release.split('.')[:2])

copyright = f'{datetime.now().year} {author}'


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

builtin_extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
]

external_extensions = [
    'sphinxarg.ext',
    'sphinx_autodoc_typehints',
    'sphinx_rtd_theme',
]

custom_extensions = [
    'autosummarycaption',
    'autosummaryfromconfig',
    'betterreplace',
    'includemdheader',
    'metadata',
    'projectmetadata',
]

extensions = builtin_extensions + external_extensions + custom_extensions

exclude_patterns = []
source_suffix = ['.rst']
templates_path = ['_templates']


add_module_names = False
autoclass_content = 'class'
autodoc_class_signature = 'mixed'
autodoc_default_options = {
    # 'exclude-members': '__weakref__',
    'inherited-members': False,
    'members': True,
    'member-order': 'bysource',
    'private-members': False,
    'special-members': '__author__, __version__, __new__',
    'undoc-members': True,
}
autodoc_docstring_signature = True
autodoc_inherit_docstrings = True
autodoc_typehints = 'signature'
autodoc_typehints_format = 'short'

autosummary_generate = True
autosummary_generate_overwrite = True
autosummary_ignore_module_all = False
autosummary_imported_members = False

napoleon_google_docstring = True
# napoleon_use_param = True

typehints_fully_qualified = False
typehints_use_signature = True
typehints_use_signature_return = True

intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ['_static']
html_theme = 'sphinx_rtd_theme'


def setup(app: Sphinx) -> None:
    """Sphinx config entry point.

    Args:
        app:
          Sphinx instance.
    """

    pass



"""Provides a template for the main ArgumentParser of the package."""

# Copyright (C) 2025  Stefano Cuizza

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


from argparse import Namespace
from collections.abc import Sequence
from datetime import datetime
from typing import override, Optional


from ...ansi import Style
from ... import metadata
from .. import ExtendableArgumentParser, SmartFormatter
from ..parents import VerboseArgumentParser


def _format_package_information() -> str:
    """Returns package information, formatted as follows:

    {package name} {package version} (*)({development status})
    Copyright (c) {year} (*){authors}.
    (*)Maintained by {maintainers}.
    (*)License: {license type}.

    (*)Project URLs:
    {website name} => {website url}.
    ...

    Fields marked with (*) will be omitted if not present in the package metadata.
    """

    authors_dict: dict[str, str] = metadata.authors()

    authors_list: list[str] = []
    for author_name, author_email in authors_dict.items():
        authors_list.append(
            ''.join((
                author_name,
                f' <{author_email}>' if author_email else '',
            ))
        )
    authors_string: str = ', '.join(authors_list)

    maintainers_list: list[str] = []
    for maintainer_name, maintainer_email in metadata.maintainers().items():
        # if a maintainer is also an author, its email address will not be
        # repeated in the maintainers section to avoid redundancy.
        maintainers_list.append(
            ''.join((maintainer_name, f' <{maintainer_email}>'))
            if (
                    maintainer_email and
                    authors_dict.get(maintainer_name, '') != maintainer_email
            )
            else maintainer_name
        )

    # if there is more than one package maintainer,
    # adds 'and' before the last name.
    maintainers_string: str = (
        maintainers_list[0] if len(maintainers_list) < 2 else
        f'{', '.join(maintainers_list[:-1])} and {maintainers_list[-1]}'
    )

    urls_list: list[str] = []
    for website_name, website_url in metadata.project_url().items():
        urls_list.append(
            f'{website_name.capitalize()} => '
            f'{Style.UNDERLINE}{website_url}{Style.RESET_UNDERLINE}.'
        )
    urls_string: str = '\n'.join(urls_list)

    dev_status: str = metadata.development_status()
    license_name: str = metadata.license_type()

    return ''.join((
        f'{metadata.package()} {metadata.version()}'
        f'{f' ({dev_status.lower()})' if dev_status else ''}\n',
        f'Copyright (c) {datetime.now().year}'
        f'{f' {authors_string}' if authors_string else ''}.\n',
        f'Maintained by {maintainers_string}.\n' if maintainers_string else '',
        f'License{'d under the ' if 'license' in license_name.lower() else ' - '}{license_name}.\n'
        if license_name else '',
        f'\nProject URLs:\n{urls_string}' if urls_string else '',
    ))


class MainArgumentParserTemplate(ExtendableArgumentParser):
    """Argument parser that adds basic CLI functionalities.

    New subparser:
        config:
            allows the user to view and modify application settings.

    New flag:
        -V, --version:
            show version information and exit.
    """

    @override
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            formatter_class=SmartFormatter,
            parents=[VerboseArgumentParser()],
            **kwargs,
        )

        self.add_argument(
            '-V', '--version',
            action='version',
            version=_format_package_information(),
        )

    @override
    def parse_args(
            self,
            args: Optional[Sequence[str]] = None,
            namespace: Optional[Namespace] = None,
    ) -> Namespace:
        return super().parse_args(args=args, namespace=namespace)



def test_author_dunder_attr(package) -> None:
    """Verifies that the package has a __author__ attribute."""

    assert hasattr(package, '__author__')


def test_doc_dunder_attr(package) -> None:
    """Verifies that the package has a __doc__ attribute."""

    assert hasattr(package, '__doc__')


def test_version_dunder_attr(package) -> None:
    """Verifies that the package has a __version__ attribute."""

    assert hasattr(package, '__version__')



"""Tokenizer that transforms a string into a series of key inputs."""

# Copyright (C) 2025  Stefano Cuizza

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


from itertools import takewhile
import re
from typing import NamedTuple

from .arduino.keyboard import SPECIAL_KEYS


type RawKeystroke = str | tuple[str, ...]
type Keystroke = str
type SleepInterval = int


FORBIDDEN_CHARACTERS: tuple[str, ...] = ('\t', '\n', '\r', '\f', '\v')
TOKENS: dict[str, str] = {
    'COMBO': r'<(?:[A-Z0-9_]+|[^A-Z])(?:\+(?:[A-Z0-9_]+|[^A-Z]))*>',
    'CHARACTER': r'(.)',
}
"""Recognized tokens."""

COMBO_KEY: str = r'([A-Z0-9_]+|[^A-Z])'


class Token(NamedTuple):
    """Represents a character of a sequence of characters."""

    type: str
    value: str


def _is_part_of_uppercase_word(c: str) -> bool:
    return (
        not c.islower() and c not in FORBIDDEN_CHARACTERS and len(c) == 1
    )


def to_keystrokes(string: str) -> list[RawKeystroke]:
    """Transforms a string of characters into a sequence of key presses.


    ==============
    Interpretation
    ==============

    Combos should have this form: <CONSTANT_or_single_key+...>.
    All other characters including line termination characters are interpreted as
    one character. All occurrences of two or more uppercase letters get enclosed
    in `KEY_CAPS_LOCK` key presses.


    ===========
    Translation
    ===========

    Keys that are supposed to be pressed at the same time are stored in tuples.
    Keys that are more than one letter long are special constants defined by the
    `Keyboard.h` library, and should be interpreted as such.
    """

    result: list[RawKeystroke] = []

    token_regex = '|'.join(f'(?P<{name}>{pattern})' for name, pattern in TOKENS.items())
    for match in re.finditer(token_regex, string, flags=re.S):
        groups = match.groups()

        match match.lastgroup:
            case 'CHARACTER':
                result.append(groups[-1])
            case 'COMBO':
                combo = groups[0].strip('<>').split('+')

                # handle abbreviated special keys.
                for i, key in enumerate(combo.copy()):
                    if len(key) > 1:
                        combo[i] = SPECIAL_KEYS.get(combo[i], combo[i])

                # flatten one character combos.
                if len(combo) > 1:
                    result.append(tuple(combo))
                else:
                    result.extend(combo)

    i: int = 0
    while i < len(result):
        key = result[i]

        if isinstance(key, tuple):
            i += 1
            continue

        # simulate the use of the `KEY_CAPS_LOCK` key with sequences of
        # uppercase characters that are longer than two letters.
        if key.isupper():
            word = tuple(takewhile(_is_part_of_uppercase_word, result[i:]))

            if len(word) > 1:
                for j in range(len(word)):
                    result[i + j] = result[i + j].lower()

                result.insert(i, SPECIAL_KEYS['CAPS_LOCK'])
                result.insert(i + len(word) + 1, SPECIAL_KEYS['CAPS_LOCK'])

                i += len(word)

        i += 1

    return result


def merge_delays(keystrokes: list[RawKeystroke]) -> list[RawKeystroke]:
    """Merges adjacent delays."""

    result: list[RawKeystroke] = []

    i: int = 0
    while i < len(keystrokes):
        raw_key = keystrokes[i]

        if isinstance(raw_key, str) or isinstance(raw_key, tuple):
            result.append(raw_key)

            i += 1
            continue

        consecutive_delays = tuple(
            takewhile(lambda j: isinstance(j, int), keystrokes[i:])
        )

        result.append(sum(consecutive_delays))
        i += len(consecutive_delays)

    return result


def keystrokes_to_arrays(keystrokes: list[RawKeystroke]) -> tuple[list[Keystroke], list[tuple[int, SleepInterval]]]:
    """Translates a sequence of raw keystrokes into an array of keystrokes and
    a sleep interval mapping.


    ========
    Examples
    ========

    The following sequence:
        [('LEFT_GUI', 'r'), 'c', 'm', 'd', '\n', 200, 'l', 's', '\n']
    Gets translated to:
        ['\0', 'LEFT_GUI', 'r', '\0', 'c', 'm', 'd', '\n', 'l', 's', '\n']
        [(6, 200)]
    """

    keys: list[Keystroke] = []
    interval_mapping: list[tuple[int, SleepInterval]] = []

    shift_correction: int = 0
    for i, raw_key in enumerate(keystrokes):
        if isinstance(raw_key, int):
            interval_mapping.append((i + shift_correction, raw_key))
            shift_correction -= 1

        elif isinstance(raw_key, str):
            keys.append(raw_key)

        elif isinstance(raw_key, tuple):
            keys.extend(('\0', *raw_key, '\0'))
            shift_correction += 3

    return keys, interval_mapping



"""Adds new functionalities to the docutils include directive.

New options:
    - :header: header
      when including a markdown file, it is possible to select only the
      contents under a certain header.
"""

from pathlib import Path
import re
from typing import Any, Callable

from docutils import io, utils
from docutils.nodes import Node
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives import misc
from sphinx.application import Sphinx


markdown_header_regex = re.compile(
    r'^\s*(?P<level>#{1,6})\s*(?P<title>[\w ]+)\s*$'
)


def apply_include_run_patch(
        func: Callable[[misc.Include], list[Node]]
) -> Callable[[misc.Include], list[Node]]:
    """Decorator function to patch the run method of the Include directive."""

    def patch(self: misc.Include) -> list[Node]:
        if (target_header_title := self.options.get('header')) is None:
            return func(self)

        if not self.state.document.settings.file_insertion_enabled:
            raise self.warning('"%s" directive disabled.' % self.name)

        current_source = self.state.document.current_source

        path = directives.path(self.arguments[0])
        if path.startswith('<') and path.endswith('>'):
            _base = self.standard_include_path
            path = path[1:-1]

        else:
            _base = Path(current_source).parent

        path = utils.relative_path(None, _base / path)

        encoding = self.options.get(
            'encoding', self.state.document.settings.input_encoding
        )
        e_handler = self.state.document.settings.input_encoding_error_handler

        try:
            include_file = io.FileInput(
                source_path=path,
                encoding=encoding,
                error_handler=e_handler,
            )
        except UnicodeEncodeError:
            raise self.severe(
                f'Problems with "{self.name}" directive path:\n'
                f'Cannot encode input file path "{path}" '
                f'(wrong locale?).'
            )
        except OSError as error:
            raise self.severe(
                f'Problems with "{self.name}" directive '
                f'path:\n{io.error_string(error)}.'
            )
        else:
            self.state.document.settings.record_dependencies.add(path)

        header_level = ''
        lines = include_file.readlines()
        self.options['start-line'],  self.options['end-line'] = 0, len(lines)
        for i, line in enumerate(lines):
            if (m := re.match(markdown_header_regex, line)) is not None:
                level, title = m.group('level'), m.group('title')

                if not header_level:
                    if title == target_header_title:
                        header_level = level
                        self.options['start-line'] = i + 1
                else:
                    if len(level) <= len(header_level):
                        self.options['end-line'] = i - 1
                        break

        return func(self)

    if hasattr(func, '__doc__'):
        patch.__doc__ = func.__doc__

    return patch


def apply_replace_patch(
        cls: misc.Include
) -> misc.Include:
    """Decorator function to patch the Include directive."""

    cls.option_spec = misc.Include.option_spec | {
        'header': directives.unchanged,
    }

    cls.run = apply_include_run_patch(cls.run)

    return cls


def setup(app: Sphinx) -> dict[str, Any]:
    """Sphinx extension entry point.

    Args:
        app:
          Sphinx instance.

    Returns:
        the metadata of the extension.
    """

    misc.Include = apply_replace_patch(misc.Include)

    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



"""Contains a collection of useful parent argument parsers.

Parent parsers should not rely on any argument parsing method (e.g. parse_args())
to parse their parsing, but should implement custom actions.
"""

# Copyright (C) 2025  Stefano Cuizza

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


from argparse import Action, ArgumentParser
from typing import override

from .. import log


class HandleVerbosityAction(Action):
    """Cumulative action that updates the application logging level
    based on the value of argument.
    """

    @override
    def __init__(self, option_strings, dest, nargs=0, **kwargs):
        super().__init__(option_strings, dest, nargs, **kwargs)

    @override
    def __call__(self, parser, namespace, values, option_string=None):
        current_value: int = getattr(namespace, self.dest, 0)
        setattr(namespace, self.dest, updated_value := current_value + 1)

        # the maximum value for verbosity is 5.
        if (verbosity := min(updated_value, 5)) > 0:
            log.set_root_logger_level(verbosity * 10)


class HandleQuietnessAction(Action):
    """Cumulative action that updates the application logging level
    based on the value of argument.
    """

    @override
    def __init__(self, option_strings, dest, nargs=0, **kwargs):
        super().__init__(option_strings, dest, nargs, **kwargs)

    @override
    def __call__(self, parser, namespace, values, option_string=None):
        current_value: int = getattr(namespace, self.dest, 0)
        setattr(namespace, self.dest, updated_value := current_value + 1)

        # the maximum value for quietness is 3.
        if (quietness := min(updated_value, 3)) > 0:
            log.set_root_logger_level((quietness + 2) * 10)


class VerboseArgumentParser(ArgumentParser):
    """Adds the ability to change verbosity through the CLI.

    New flags:
        -v, --verbose:
            set the verbosity level. This option is additive,
            and can be used up to 5 times.
        -q, --quiet:
            give less output. This option is additive,
            and can be used up to 3 times.
    """

    @override
    def __init__(self, *args, **kwargs):
        kwargs['add_help'] = False
        super().__init__(*args, **kwargs)

        self.add_argument(
            '-q', '--quiet',
            action=HandleQuietnessAction,
            default=0,
            dest='quiet',
            help=(
                'give less output. option is additive, and can '
                'be used up to 3 times (warning/error/critical)'
            ),
            required=False,
        )

        self.add_argument(
            '-v', '--verbose',
            action=HandleVerbosityAction,
            default=0,
            dest='verbosity',
            help=(
                'set the verbosity level. option is additive, '
                'and can be used up to 5 times '
                '(debug/info/warning/error/critical). '
                '(default: info)'
            ),
            required=False,
        )



"""Argument parsing module.

Contains extendable ArgumentParser subclasses, formatters and helper functions.
"""

# Copyright (C) 2025  Stefano Cuizza

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


from abc import ABC, abstractmethod
from argparse import (
    ArgumentParser,
    HelpFormatter,
    Namespace,
)
from collections.abc import Sequence
import logging
from typing import Optional, override


logger = logging.getLogger(__name__)


# ruamel (2020, March 30). Argparse improvements. SourceForge.
# https://sourceforge.net/projects/ruamel-std-argparse/.
class SmartFormatter(HelpFormatter):
    """Enables RawTextHelpFormatter for help messages that begin with `R|`
    and preserves line breaks in the `version` action.
    """

    def _fill_text(self, text, width, indent):
        return ''.join([indent + line for line in text.splitlines(True)])

    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()

        return HelpFormatter._split_lines(self, text, width)


class ExtendableArgumentParser(ABC, ArgumentParser):
    """Argument parser with extendable functionality.

    Implement two private methods to extend functionality:
     - `_extend_arguments()` to add new arguments.
     - `_extend_subparsers()` to add new subparsers.
    """

    @override
    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self._subparsers = None
        self._parents: list[ArgumentParser] = kwargs.get('parents', [])

        self._extend_arguments()
        self._extend_subparsers()

    @abstractmethod
    def _extend_arguments(self) -> None:
        """Adds new arguments."""

    @abstractmethod
    def _extend_subparsers(self) -> None:
        """Adds new subparsers."""

    @override
    def parse_args(
            self,
            args: Optional[Sequence[str]] = None,
            namespace: Optional[Namespace] = None,
    ) -> Namespace:
        namespace = super().parse_args(args=args, namespace=namespace)

        logger.debug('parsing: %r.', vars(namespace))
        return namespace

    @property
    def subparsers(self):
        """The argument parser subparsers.

        Subparsers are loaded lazily to avoid unnecessary overhead.
        """

        if not self._subparsers:
            # enable subparsers only if needed, to avoid weird artifacts in the help message.
            self._subparsers = self.add_subparsers(parser_class=ArgumentParser, dest='subparser')

        return self._subparsers



"""Provides easy access to the package metadata."""

# Copyright (C) 2025  Stefano Cuizza

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.


import email.policy
from email import message_from_string
import logging
from importlib.metadata import (
    metadata,
    PackageMetadata,
    PackageNotFoundError,
)
from operator import methodcaller
from typing import Literal


logger = logging.getLogger(__name__)


def package() -> str:
    """Returns the package name."""

    return __package__.split('.', maxsplit=1)[0]


package_name = package()

try:
    package_metadata: PackageMetadata = metadata(package_name)
except PackageNotFoundError:
    logger.critical('%r package not found.', package_name)
    raise


def version() -> str:
    """Returns the package version."""

    return package_metadata.get_all('Version', [''])[0]


def summary() -> str:
    """Returns the package summary."""

    return package_metadata.get_all('Summary', [''])[0]


def readme() -> str:
    """Returns the contents of the package README.md file."""

    return package_metadata.get_all('Description', [''])[0]


def license() -> str:
    """Returns the package license."""

    return package_metadata.get_all('License', [''])[0]


def authors() -> dict[str, str]:
    """Returns a dictionary mapping package authors to
    their corresponding email addresses.
    """

    return _field_with_email('Author')


def maintainers() -> dict[str, str]:
    """Returns a dictionary mapping package maintainers to
    their corresponding email addresses.
    """

    return _field_with_email('Maintainer')


def classifiers() -> list[str]:
    """Returns a list containing package classifiers.

    For a comprehensive list of classifiers see: https://pypi.org/classifiers.
    """

    return package_metadata.get_all('Classifier', [])


def development_status() -> str | None:
    """Returns the package development status.

    The status is extracted from the "Development Status" classifier (if present).

    Common possible return values:
        - Planning
        - Pre-Alpha
        - Alpha
        - Beta
        - Production/Stable
        - Mature
        - Inactive
    """

    for classifier_type, classifier_body in map(
            methodcaller('split', ' :: ', 1),
            classifiers(),
    ):
        if classifier_type == 'Development Status':
            _status_number, status_literal = classifier_body.split(' - ')

            return status_literal

    return None


def license_type() -> str | None:
    """Returns the package license type.

    The package license type is extracted from the "License" classifier (if present).
    """

    for classifier_type, classifier_body in map(
            methodcaller('split', ' :: ', 1),
            classifiers(),
    ):
        if classifier_type == 'License':
            return classifier_body.split(' :: ')[-1]

    return None


def project_url() -> dict[str, str]:
    """Returns a dictionary mapping the project websites
    with their corresponding urls.
    """

    urls: dict[str, str] = {}
    project_urls: list[str] = (
        package_metadata.get_all('Project-URL', [])
    )

    # Raymond Hettinger (2011, December 11). python map (...). StackOverflow.
    # https://stackoverflow.com/a/8461254.
    for website_name, website_url in map(
            methodcaller('split', ', '),
            project_urls,
    ):
        urls[website_name] = website_url

    return urls


def _field_with_email(field: Literal['Author', 'Maintainer']) -> dict[str, str]:
    """Returns a dictionary mapping names to their corresponding email addresses
    from a specific metadata field.

    Args:
        field:
            the target metadata field.
    """

    field_metadata: dict[str, str] = {}

    # for more information on the fields structure see:
    # https://packaging.python.org/en/latest/specifications/declaring-project-metadata/#authors-maintainers.
    name_fields: list[str] = (
        package_metadata.get_all(field, [])
    )
    fields_with_emails: list[str] = (
        package_metadata.get_all(f'{field}-email', [])
    )

    for name_field in name_fields:
        field_metadata[name_field] = ''

    for email_field in fields_with_emails:
        # sinoroc (2023, March 21). importlib.metadata (...). StackOverflow.
        # https://stackoverflow.com/a/75803208.
        email_field_message = message_from_string(
            f'To: {email_field}',
            policy=email.policy.default,
        )

        for entry in email_field_message['to'].addresses:
            if not (display_name := entry.display_name):
                # guess the name from the username portion of the email address.
                display_name: str = ' '.join(
                        map(str.capitalize, entry.username.split('.'))
                    )

            field_metadata[display_name] = entry.addr_spec

    return field_metadata
